///
/// Copyright (c) 2016 Dropbox, Inc. All rights reserved.
///
/// Auto-generated by Stone, do not modify.
///

import Foundation

public class {{ class_name }}: DropboxTransportClientOwning {
    public var client: {{ transport_client_name }}

    {% for var, type in namespace_fields %}
    /// Routes within the {{ var }} namespace. See {{ type }}Routes for details.
    public var {{ var }}: {{ type }}Routes!
    {% endfor %}

    public required init(client: {{ transport_client_name }}) {
        self.client = client

        {% for var, type in namespace_fields %}
        self.{{ var }} = {{ type }}Routes(client: client)
        {% endfor %}
    }
}

