///
/// Copyright (c) 2022 Dropbox, Inc. All rights reserved.
///
/// Auto-generated by Stone, do not modify.
///

import Foundation
import SwiftyDropbox

/// Objective-C compatible {{ class_name }}.
/// For Swift see {{ class_name }}.
@objc
public class DBX{{ class_name }}: NSObject {
    let swift: {{ class_name }}

    {% for var, type in namespace_fields %}
    /// Routes within the {{ var }} namespace. See DB{{ type }}Routes for details.
    @objc
    public var {{ var }}: DBX{{ type }}Routes!
    {% endfor %}

    @objc
    public convenience init(client: DBXDropboxTransportClient) {
        self.init(swiftClient: client.swift)
    }

    public init(swiftClient: DropboxTransportClient) {
        self.swift = {{ class_name }}(client: swiftClient)

        {% for var, type in namespace_fields %}
        self.{{ var }} = DBX{{ type }}Routes(swift: swift.{{ var }})
        {% endfor %}
    }
}

