///
/// Copyright (c) 2022 Dropbox, Inc. All rights reserved.
///
/// Auto-generated by Stone, do not modify.
///

import Foundation
import SwiftyDropbox

extension {{ class_name }} {
    var objc: DBXRequest {
        {% for route_args_data in background_objc_routes %}
        {% set namespace = route_args_data[0] %}
        {% set route = route_args_data[1] %}
        {% set args_data = route_args_data[2] %}
        if case .{{ fmt_func_namespace(route.name, route.version, namespace.name) }}(let swift) = self {
            return {{ fmt_route_objc_class(namespace, route, args_data) }}(swift: swift)
        }
        {% endfor %}
        else {
            fatalError("For Obj-C compatibility, add this route to the Objective-C compatibility module allow-list")
        }
    }
}

